/* 
 * Birdstep Database Engine
 *
 * Copyright (C) 1999, 2000 by Birdstep Technology AS 
 * All rights reserved.
 *
 * bde.h
 *
 * Birdstep Database Engine Application Programming Interface header file.
 *
 * $Id: bde.h,v 1.1.2.1 2000/09/21 09:06:20 mlj Exp $
 *
 */

#ifndef __BDE_H__
#define __BDE_H__

/* ============================ */
/* Includes and general defines */
/* ============================ */

/* EPOC stuff. */
#ifdef __SYMBIAN32__
#undef WIN32
#undef _WIN32
#endif

/* Windows/mingw32 stuff. */
#if defined(_WIN32) && !defined(UNDER_CE) && !defined(__MINGW32__)
#pragma warning(push,3)
#endif

#ifdef _WIN32
#include <wtypes.h>
#include <tchar.h>
#endif /* _WIN32 */

#if defined(_WIN32) && !defined(UNDER_CE) && !defined(__MINGW32__)
#pragma warning(pop)
#endif

/* VxWorks stuff. */
#ifdef __vxworks__
#include <stdlib.h>
#endif /* __vxworks__ */

/* UN*X stuff. */
#if defined(unix) && !defined(__unix__)
#define __unix__
#endif /* unix */

/* Global API functions */
#if defined(WIN32) || defined(_WIN32)
#ifdef IBAPI_STATIC
#define IBCALL
#else
#ifdef IBAPI_EXPORTS
#define IBCALL __declspec(dllexport)
#else
#define IBCALL __declspec(dllimport)
#endif
#endif
#elif defined(__SYMBIAN32__)
#define IBCALL __declspec(dllexport)
#else
#define IBCALL
#endif

/* ================ */
/* Type definitions */
/* ================ */

/* Type ranges. */
#define IBU8_MIN                        0
#define IBU8_MAX                        255
#define IBU16_MIN                       0
#define IBU16_MAX                       65535
#define IBU32_MIN                       0
#define IBU32_MAX                       4294967295U
#define IBU64_MIN                       0
#define IBU64_MAX                       18446744073709551615
#define IBS8_MIN                        (-128)
#define IBS8_MAX                        127
#define IBS16_MIN                       (-32768)
#define IBS16_MAX                       32767
#define IBS32_MIN                       (-2147483647L)
#define IBS32_MAX                       2147483647
#define IBS64_MIN                       (-9223372036854775808)
#define IBS64_MAX                       9223372036854775807

#define IBUCHAR_MIN                     IBU8_MIN
#define IBUCHAR_MAX                     IBU8_MAX
#define IBUSHORT_MIN                    IBU16_MIN
#define IBUSHORT_MAX                    IBU16_MAX
#define IBULONG_MIN                     IBU32_MIN
#define IBULONG_MAX                     IBU32_MAX
#define IBCHAR_MIN                      IBS8_MIN
#define IBCHAR_MAX                      IBS8_MAX
#define IBSHORT_MIN                     IBS16_MIN
#define IBSHORT_MAX                     IBS16_MAX
#define IBLONG_MIN                      IBS32_MIN
#define IBLONG_MAX                      IBS32_MAX
#define IBDBADDR_MIN                    IBULONG_MIN
#define IBDBADDR_MAX                    IBULONG_MAX

/* Base types */
enum {
    IBBASE_TYPE_MIN                 = 1,
    
    /* Simple base types. */
    IBBASE_TYPE_NONE                = 0,
    IBBASE_TYPE_UCHAR               = 1, 
    IBBASE_TYPE_CHAR                = 2,
    IBBASE_TYPE_USHORT              = 3, 
    IBBASE_TYPE_SHORT               = 4, 
    IBBASE_TYPE_ULONG               = 5,
    IBBASE_TYPE_DBADDR              = 5,
    IBBASE_TYPE_LONG                = 6, 
    IBBASE_TYPE_TEXT                = 8, 
    IBBASE_TYPE_UNICODE             = 9, 
    IBBASE_TYPE_BLOB                = 10,

    IBBASE_TYPE_MAX                 = 10,

    IBBASE_TYPE_NUM                 = IBBASE_TYPE_MAX + 1
};

/* Storage types. */
#define IBSTORAGE_TYPE_MASK             0x01
#define IBSTORAGE_TYPE_PRIMARY          0x00
#define IBSTORAGE_TYPE_SECONDARY        0x01

/* Fixed size types. */
#if !defined(__GNUC__) || __GNUC__ < 2 || __GNUC_MINOR__ < 7

#if defined(__palmos__) || defined(__PALMOS_TRAPS__)
#include <PalmOS.h>
typedef UInt8 ibu8;
typedef UInt16 ibu16;
typedef UInt32 ibu32;
typedef Char ibs8;
typedef Int16 ibs16;
typedef Int32 ibs32;
typedef Char ibtchar;
#endif /* __palmos__ */

#if defined(__unix__)
/* #include <sys/types.h> */
#include <stdint.h>
typedef int8_t   ibs8;
typedef int16_t  ibs16;
typedef int32_t  ibs32;
typedef uint8_t  ibu8;
typedef uint16_t ibu16;
typedef uint32_t ibu32;
typedef char ibtchar;
#endif /* __unix__ */

#ifdef __SYMBIAN32__
typedef unsigned char ibu8;
typedef unsigned short int ibu16;
typedef unsigned long int ibu32;
typedef signed char ibs8;
typedef short int ibs16;
typedef long int ibs32;
typedef char ibtchar;
#endif /* __SYMBIAN32__ */

#if defined(_WIN32)
typedef unsigned char ibu8;
typedef unsigned short ibu16;
typedef unsigned long ibu32;
typedef char ibs8;
typedef short ibs16;
typedef long ibs32;
typedef __int64 ibs64;
typedef TCHAR ibtchar;
#endif /* _WIN32 */

#else

/* This will do for now, but expect it to break some day... [mlj] */
typedef signed int ibs8 __attribute__ ((__mode__(__QI__)));
typedef signed int ibs16 __attribute__ ((__mode__(__HI__)));
typedef signed int ibs32 __attribute__ ((__mode__(__SI__)));
typedef signed int ibs64 __attribute__ ((__mode__(__DI__)));
typedef unsigned int ibu8 __attribute__ ((__mode__(__QI__)));
typedef unsigned int ibu16 __attribute__ ((__mode__(__HI__)));
typedef unsigned int ibu32 __attribute__ ((__mode__(__SI__)));
typedef unsigned int ibu64 __attribute__ ((__mode__(__DI__)));
typedef char ibtchar;
#endif /* __GNUC__ || __GNUC__ < 2 || __GNUC_MINOR__ < 7 */

/* System types. */
typedef struct ibcursor *ibcursorhandle;   /* Cursor handle */
typedef struct ibdct *ibdcthandle;         /* Dictionary handle */
typedef ibu32 ibdcttypehandle;             /* Dictionary type handle */
typedef struct ibdb *ibdbhandle;           /* Database handle */
typedef struct ibres *ibreshandle;         /* Result set handle */
typedef struct ibquery *ibqueryhandle;     /* Query handle */
typedef struct ibenv *ibenvhandle;         /* Environment handle */
typedef struct ibreslist *ibreslisthandle; /* Result list handle */

#define IBDCT_USER_TYPE_MAX             IBU16_MAX
typedef int ibutype;                    /* User type */
typedef int ibbtype;                    /* Base type */
typedef int ibstype;                    /* Storage type */

typedef ibu8 ibuchar;                   /* Database type "UCHAR" */
typedef ibu16 ibushort;                 /* Database type "USHORT" */
typedef ibu32 ibulong;                  /* Database type "ULONG" */
typedef ibs8 ibchar;                    /* Database type "CHAR" */
typedef ibs16 ibshort;                  /* Database type "SHORT" */
typedef ibs32 iblong;                   /* Database type "LONG" */
typedef ibulong ibdbaddr;               /* Database type "DBADDR"
                                           database address */

typedef void ibechar;                   /* Coded character */
typedef void ibutf8;                    /* UTF-8 coded character */
typedef ibu32 ibobjsize;                /* Size of a data buffer or
                                           data object in bytes. */ 
typedef unsigned int ibbufsize;         /* Size of internal buffers. */
typedef unsigned int ibbufnum;          /* Number of internal buffers. */
typedef unsigned long ibir;             /* Index into or size of a
                                           record. */ 
typedef unsigned long ibrl;             /* Index into or size of a
                                           result list. */ 

/* ======= */
/* Defines */
/* ======= */

/* Error codes. */
typedef enum {
    /* Error codes without additional information. */
    IBERR_NONE = 0, 

    IBERR_NOT_IMPLEMENTED,

    IBERR_DB_INVALID_HANDLE, IBERR_DB_ILLEGAL_FLAGS,
    IBERR_DB_INVALID_PARAMETER, IBERR_DB_DRIVER_UNAVAILABLE, 

    IBERR_CURSOR_INVALID_HANDLE, IBERR_CURSOR_NO_CURRENT_RECORD,
    IBERR_CURSOR_NO_CURRENT_FIELD, IBERR_CURSOR_INVALID_IR_INDEX,
    IBERR_CURSOR_SUPPLIED_FETCH_BUFFER_TOO_SMALL, 
    IBERR_CURSOR_UNABLE_TO_DELETE_OWNER,

    IBERR_DCT_INVALID_HANDLE, IBERR_DCT_TYPE_NAME_NOT_FOUND,
    IBERR_DCT_TYPE_NOT_FOUND, IBERR_DCT_INCONSISTENT,
    IBERR_DCT_TYPE_NAME_ALREADY_DEFINED, IBERR_DCT_TYPE_TOO_MANY,

    IBERR_RES_INVALID_HANDLE,

    IBERR_RES_LIST_INVALID_INDEX, IBERR_RES_LIST_INVALID_HANDLE, 

    IBERR_QUERY_INVALID_HANDLE, IBERR_QUERY_INVALID_EXPRESSION,
    IBERR_QUERY_INFINITE_RECURSION, IBERR_QUERY_INVALID_PARAMETERS,

    IBERR_SET_INVALID_BINARY_OPERATOR,

    IBERR_ENV_INVALID_HANDLE,

    IBERR_SYS_BASE_TYPE_INVALID, 
    IBERR_SYS_DATATOOBIG, 
    IBERR_SYS_ILLEGAL_DB_FILE, 
    IBERR_SYS_ILLEGAL_DB_VERSION, 

    /* Error codes accompanied by additional errno information. */
    IBERR_MALLOC = 0x1000, IBERR_REALLOC, IBERR_OPEN,
    IBERR_STAT, IBERR_CLOSE, IBERR_MUNMAP, IBERR_MMAP, IBERR_MREMAP,
    IBERR_FTRUNCATE, IBERR_MSYNC,
    IBERR_FCLOSE, IBERR_FOPEN, IBERR_FTELL, IBERR_FREAD, IBERR_FSEEK,
    IBERR_FFLUSH, IBERR_CALLOC, IBERR_FSTAT, IBERR_FSYNC, IBERR_WRITE,
    IBERR_LSEEK, IBERR_READ, IBERR_FWRITE, IBERR_FORK, IBERR_SETSID,
    IBERR_SOCKET, IBERR_SETSOCKOPT, IBERR_BIND, IBERR_LISTEN,
    IBERR_ACCEPT, IBERR_GETHOSTBYNAME, IBERR_CONNECT,

    /* Error codes accompanied by additional Windows error information. */
    IBERR_CREATEFILE = 0x2000, IBERR_FLUSHFILEBUFFERS,
    IBERR_CLOSEHANDLE, IBERR_SETFILEPOINTER, IBERR_READFILE,
    IBERR_WRITEFILE, IBERR_GETFILESIZE, IBERR_GETFILESIZE_FILE4GB,
    IBERR_MAPVIEWOFFILE, IBERR_CREATEFILEMAPPING,
    IBERR_UNMAPVIEWOFFILE, IBERR_FLUSHVIEWOFFILE,
    IBERR_GETFILEATTRIBUTES, 

    /* Error codes for PalmOS. DmGetLastErr() will proivde details */
    IBERR_CREATEDATABASE = 0x3000, IBERR_OPENDATABASE, 
    IBERR_CLOSEDATABASE, IBERR_FINDDATABASE, IBERR_DELETEDATABASE,
    IBERR_NEWRECORD, IBERR_RESIZERECORD, IBERR_GETRECORD, 
    IBERR_RELEASERECORD, IBERR_QUERYRECORD, 
    IBERR_MEMWRITE, IBERR_MEMSET,

    /* Error codes for PalmOS without additional information. */
    IBERR_MEMPTRNEW = 0x3500, IBERR_MEMPTRFREE, 
    IBERR_MEMHANDLELOCK, IBERR_MEMHANDLEUNLOCK, IBERR_MEMMOVE,

    IBERR_MAX 
} ibret;

/* Error code masks. */
#define IBERR_MASK_CODE                 0x0fff
#define IBERR_MASK_FLAG_ERRNO           0x1000
#define IBERR_MASK_FLAG_WINERR          0x2000

/* Encodings. */
typedef enum {
    IBENC_US_ASCII = 0,                 /* 7-bit US-ASCII. */

    IBENC_ISO_8859_1,                   /* ISO-8859-1 or Latin-1. */
    IBENC_LATIN_1 = IBENC_ISO_8859_1,

    IBENC_UTF_8,                        /* UTF-8. */
    IBENC_UTF_2 = IBENC_UTF_8,
    IBENC_UTF_FSS = IBENC_UTF_8,

    IBENC_UTF_16LE,                     /* UTF-16, little-endian
                                           version. */

    IBENC_UTF_16BE,                     /* UTF-16, big-endian */
    IBENC_UTF_16 = IBENC_UTF_16BE,      /* version. */

    IBENC_UCS_2LE, 

    IBENC_UCS_2BE,
    IBENC_UCS_2 = IBENC_UCS_2BE,
    IBENC_ISO_10646_UCS_2 = IBENC_UCS_2BE,

    IBENC_UCS_4,
    IBENC_ISO_10646 = IBENC_UCS_4,
    IBENC_ISO_10646_UCS_4 = IBENC_UCS_4
} ibenc;

/* Determine native charater set */
#if defined(UNDER_CE) || (defined(_WIN32) && defined(_UNICODE))
#define IBENC_NATIVE_CHARSET IBENC_UTF_16LE
#else
#define IBENC_NATIVE_CHARSET IBENC_US_ASCII
#endif

#define IBENC_DEFAULT        IBENC_NATIVE_CHARSET

struct ibdtparams {
    ibenc encoding;                     /* Data encoding. Only
                                           relevant for stings. */
};
    
/* Unary set operators. */
typedef enum {
    IBSET_UNOP_NONE,                    /* No operation. */
    IBSET_UNOP_COMP                     /* Complement of a result
                                           set. */ 
} ibsetunop;

/* Binary set operator. */
typedef enum {
    IBSET_BINOP_UNION,                  /* The union of two result
                                           sets. */
    IBSET_BINOP_INTERSECT,              /* The intersection of two
                                           result sets. */
    IBSET_BINOP_DIFF,                   /* The difference between two
                                           result sets. */
    IBSET_BINOP_SYM_DIFF                /* The symmetric difference
                                           between two result sets. */
} ibsetbinop;

/* Flags for the ibdbparams structure. Please note that not all
   combinations of these flags are supported on all achitectures.  */

/* Masks. */
#define IBDBFLAG_OPEN_MASK              0x0000000F
#define IBDBFLAG_ACCESS_MASK            0x00000030
#define IBDBFLAG_SHARE_MASK             0x000000C0
#define IBDBFLAG_MASK_MODE              (IBDBFLAG_OPEN_MASK | IBDBFLAG_ACCESS_MASK | IBDBFLAG_SHARE_MASK)

/* Open mode. */
#define IBDBFLAG_OPEN                   0x00000001
#define IBDBFLAG_CREATE                 0x00000002
#define IBDBFLAG_TRUNCATE               0x00000004
#define IBDBFLAG_OPEN_EXISTING          IBDBFLAG_OPEN
#define IBDBFLAG_OPEN_CREATE            (IBDBFLAG_OPEN | IBDBFLAG_CREATE)
#define IBDBFLAG_CREATE_ALWAYS          (IBDBFLAG_OPEN | IBDBFLAG_CREATE | IBDBFLAG_TRUNCATE)
#define IBDBFLAG_TRUNCATE_EXISTING      (IBDBFLAG_OPEN | IBDBFLAG_TRUNCATE)

/* Access mode. */
#define IBDBFLAG_ACCESS_READ            0x00000010
#define IBDBFLAG_ACCESS_WRITE           0x00000020
#define IBDBFLAG_ACCESS_READWRITE       (IBDBFLAG_ACCESS_READ | IBDBFLAG_ACCESS_WRITE)

/* Share mode. */
#define IBDBFLAG_SHARE_EXCLUSIVE        0x00000000
#define IBDBFLAG_SHARE_READ             0x00000040
#define IBDBFLAG_SHARE_WRITE            0x00000080
#define IBDBFLAG_SHARE_READWRITE        0x000000C0

/* Short-hand for some commonly used file modes. */

/* Always create file. Overwrite any existing file. Create in
   read/write mode. */ 
#define IBDBFLAG_CREATE_ALWAYS_RW       (IBDBFLAG_CREATE_ALWAYS | IBDBFLAG_ACCESS_READWRITE)
/* Open file if it exists, create if no file exists. Open or create in
   read/write mode. */ 
#define IBDBFLAG_OPEN_RW                (IBDBFLAG_OPEN_CREATE | IBDBFLAG_ACCESS_READWRITE)
/* Always open file. Open in read/write mode. */
#define IBDBFLAG_OPEN_ALWAYS_RW         (IBDBFLAG_OPEN_EXISTING | IBDBFLAG_ACCESS_READWRITE)
/* Always open file. Open in read only mode. */
#define IBDBFLAG_OPEN_ALWAYS_RO         (IBDBFLAG_OPEN_EXISTING | IBDBFLAG_ACCESS_READ)

typedef enum {
    IBDB_DRIVER_DEFAULT = 0,            /* Default driver. */
    IBDB_DRIVER_MAP,                    /* File mapping driver. */
    IBDB_DRIVER_PAGING,                 /* Paging driver. */
    IBDB_DRIVER_RAMDISK,                /* Ramdisk driver. */
    IBDB_DRIVER_INMEMORY,               /* In-memory driver. */ 
    IBDB_DRIVER_CUSTOM                  /* Custom driver. */
} ibdbdriver;

struct ibdbparams {
    unsigned long flags;                /* Database file flags. */

    unsigned long initial_size;         /* Initial size of dynamically
                                           allocated database space in
                                           bytes. Must be a power of two. 
                                           0 chooses the default
                                           value. */
                                           
    unsigned long expansion_size;       /* Number of bytes to expand
                                           the dynamically allocated
                                           database space with when
                                           needed. Must be a power of
                                           two. 0 chooses the
                                           default value. If using the
                                           paging driver, the expansion
                                           size must be a multiple of the
                                           buffer size. */

    ibdbdriver driver;                  /* Which driver to
                                           use. If IBDB_DRIVER_DEFAULT,
                                           will use the default
                                           driver. */

    /* Driver specific parameters: Paging driver. */

    ibbufsize bufsize;                  /* Size of paging
                                           buffers. Must be a power of 
                                           two. */

    ibbufnum bufnum;                    /* Number of buffers. */
};

struct ibenvparams {
    /* Configuration information. */
    int local_filesystem;               /* Non-zero if using a local
                                           file system. Zero if using
                                           a network connection. */

    /* Networking information. */
    const char *hostname;               /* Name of host or IP address
                                           in quad-notation. */ 
    unsigned int port;                  /* Port to connect to or 0 if
                                           standard port. */

    /* Authentication information. */
    const char *username;               /* User name. NULL for guest
                                           connection. */
    const char *password;               /* Password. NULL for no
                                           password. */
};

struct ibsysinfo {
    /* Version information. */
    int major_version;                  /* Major version. */
    int minor_version;                  /* Minor version. */
    int patchlevel;                     /* Patch-level. */

    /* Driver information. */
    ibdbdriver default_driver;          /* Default driver. */
};

typedef enum {
    IBQUERY_NODE_TYPE_INVALID,
    IBQUERY_NODE_TYPE_BINOP,            /* A binary operator, i.e. a set operator. */
    IBQUERY_NODE_TYPE_UNOP              /* A unary operator, i.e. a set criterium. */
} ibqueryparamstype;

typedef enum {
    IBQUERY_CRITERIUM_LEVEL_FIELD,      /* Queries are executed on field level. */
    IBQUERY_CRITERIUM_LEVEL_RECORD      /* Queries are executed on record level. */
} ibquerycriteriumlevel;

typedef enum {
    /* A range search with lower and upper boundary. */
    IBQUERY_CRITERIUM_TYPE_RANGE,    

    /* An exact search. */
    IBQUERY_CRITERIUM_TYPE_EXACT,        

    /* A search over the complete data range of the type. */
    IBQUERY_CRITERIUM_TYPE_COMPLETE,     

    /* A search for items which contain the data item. */
    IBQUERY_CRITERIUM_TYPE_LIKE
} ibquerycriteriumtype;

struct ibqueryparams {
    ibqueryparamstype type;              /* Type of expression node. */

    union _data {

        struct _operation {               /* Set operator. */
            ibsetbinop op;               /* Binary set operator. */

            ibqueryhandle left_expr;     /* Handle of left query operand. */
            ibqueryhandle right_expr;    /* Handle of right query
                                            operand. */ 
        } operation;

        struct _criterium {              /* Set criterium. */
            ibquerycriteriumtype type;   /* Type of criterium. */
            ibquerycriteriumlevel level; /* Record or field level
                                            query. */ 
            ibsetunop op;                /* Unary set operator. */

            void *data1;                 /* Pointer to data item 1. */
            void *data2;                 /* Pointer to data item 2. */
            ibenc enc1;                  /* Encoding of data item 1. */
            ibenc enc2;                  /* Encoding of data item 2. */
            ibdcttypehandle data_type;   /* Type of data. */
        } criterium;
    } data;
};

typedef enum {
    IBQUERY_DESTROY_NON_RECURSIVE,       /* Destroy query handles
                                            non-recursively. */ 
    IBQUERY_DESTROY_RECURSIVE            /* Destroy query handles
                                            recursively. */ 
} ibquerydestroyflags;

typedef enum {
    IBRESLIST_SORT_ASCENDING,            /* Sort in ascending order. */
    IBRESLIST_SORT_DESCENDING,           /* Sort in descending order. */
    IBRESLIST_SORT_CALLBACK              /* Sort using callback. */
} ibreslistsorttype;

struct ibreslistsortstep {
    /* Type of sorting operation. */
    ibreslistsorttype type;         
     
    /* Field index (relative to the start of the record) to sort by. */
    ibir field_number;                   

    /* Optional call-back function used when not sorting in ascending
       or descending order. */
    int (*sorter)(ibdcttypehandle type1, const void *data1,
                  ibdcttypehandle type2, const void *data2);
};

/* Macros for generating types when not relying on the dictionary. */
#define IBDCT_MAKE_TYPE_HANDLE(b, s, u) ( ((b) & 0xff) | \
                                          (((s) & 0xff) << 8) | \
                                          (((u) & 0xffff) << 16) )
#define IBDCT_MAKE_BTYPE(handle)        ( (ibbtype)((handle) & 0xff) )
#define IBDCT_MAKE_STYPE(handle)        ( (ibstype)(((handle >> 8)) & 0xff) )
#define IBDCT_MAKE_UTYPE(handle)        ( (ibutype)(((handle >> 16)) & 0xffff) )

/* Function prototypes */
/* =================== */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

    /* General functions. */
    /* ================== */

    /** Return information about the system.
     * Returns information about the system and its capabilities.
     *
     * @param info                Pointer to a user-allocated
     *                            structure which will be filled in by
     *                            the function. 
     * @return                    Standard error code.
     */
    ibret IBCALL IBSysGetInfo(struct ibsysinfo *info);

    /* Environment functions. */
    /* ====================== */

    /** Create an environment handle.
     * Creates an environment handle and sets some environment
     * parameters. 
     *
     * @param env                 Location for the returned
     *                            environment handle.
     * @param params              Pointer to a structure containing
     *                            information about the desired
     *                            environment. 
     * @return                    Standard error code.
     */
    ibret IBCALL IBEnvCreate(ibenvhandle *env, 
                             struct ibenvparams *params);

    /** Destroy an environment handle.
     * Destroys an environment handle.
     *
     * @param env                 Environment handle to destroy.
     * @return                    Standard error code.
     */
    ibret IBCALL IBEnvDestroy(ibenvhandle env);

    /* Database file management functions. */
    /* =================================== */

    /** Open an existing database or create a new database.
     * Opens an existing database or creates a new database and returns a
     * database handle associated with the database file. 
     *
     * The file is opened or created depending on the settings in the
     * ibdbparams structure. If the flags member of the ibdbparams
     * structure is set to IBDBFLAG_CREATE_ALWAYS_RW, a new database file
     * will always be created. If a file already exists, the file will be
     * overwritten.  
     *
     * If the flag is set to IBDBFLAG_OPEN_ALWAYS_RW or
     * IBDBFLAGS_OPEN_ALWAYS_RO,  the file will always be
     * opened. If no database file exists, the function will fail and
     * return an operating system specific error, e.g. IBERR_FOPEN,
     * IBERR_OPEN.
     *
     * If the flag is set to IBDBFLAG_OPEN, an existing database file will
     * be opened, and a new database file will be created if no file
     * exists. 
     *
     * The bufnum member of the ibdbparams structure specifies the number
     * of internal buffers to be used by the input-output system. This value
     * is ignored on some systems (in general all systems which support
     * file mapping), but should be set to a valid value for portability.
     *
     * The 'path' parameter specifies the name of the database
     * file. The name may be supplied including a relative or an
     * absolute path with or without file name extension, which will
     * be used without modification to access the file.
     *
     * @param db                  Location for the returned database
     *                            handle. 
     * @param env                 Environment to use.
     * @param path                Path and file name of database file.
     * @param params              Pointer to a structure of database parameters.
     * @return                    Standard error code.
     * @see                       IBDbCommit
     * @see                       IBDbClose
     * @see                       IBDbGetParams
     */
    ibret IBCALL IBDbCreate(ibdbhandle *db, ibenvhandle env, 
                            const ibtchar *path,
                            struct ibdbparams *params);
    
    /** Commit database.
     * Commits the database file, i.e. flushes all database buffers and
     * synchronises all memory regions if needed. When the function
     * returns the database image is valid and consistent. Note that any 
     * subsequent database manipulations may invalidate the image.
     *
     * @param db                  Database handle.
     * @return                    Standard error code.
     * @see                       IBDbCreate
     * @see                       IBDbClose
     * @see                       IBDbGetParams
     */
    ibret IBCALL IBDbCommit(ibdbhandle db);

    /** Close database.
     * Closes the database file, i.e. flushes all database buffers and
     * releases all memory regions mapped to the database file and
     * destroys the database handle. Also closes all file handles and
     * other operating system specific handles associated with the file
     * and its memory regions.
     *
     * @param db                  Database handle.
     * @return                    Standard error code.
     * @see                       IBDbCreate
     * @see                       IBDbCommit
     * @see                       IBDbGetParams
     */
    ibret IBCALL IBDbClose(ibdbhandle db);

    /** Get database parameters.
     * Fills in a ibdbparams structure with information about the database
     * file associated with a database handle.
     *
     * @param db                  Database handle.
     * @param params              Pointer to a structure which will be
     *                            filled in with the database parameters.
     * @return                    Standard error code.
     * @see                       IBDbCreate
     * @see                       IBDbCommit
     * @see                       IBDbClose
     * @see                       IBDbGetName
     */
    ibret IBCALL IBDbGetParams(ibdbhandle db, struct ibdbparams *params);

    /** Get name of database file.
     * Returns the name of the database file in a user-supplied
     * buffer. The returned name is the same as the name supplied when
     * the database was created or opened, i.e. if a full path was
     * supplied, the path is returned too.  
     *
     * @param db                  Database handle.
     * @param path                User-supplied buffer for the name of
     *                            the database.
     * @param n                   The size of the user-supplied buffer
     *                            in bytes.
     * @return                    Standard error code.
     * @see                       IBDbGetParams
     */
    ibret IBCALL IBDbGetName(ibdbhandle db, ibtchar *path, ibobjsize n);

    /* Field and record manipulating functions. */
    /* ======================================== */

    /** Create a new cursor handle.
     * Creates a new cursor handle and binds the cursor handle to a database
     * handle. The effect of this is that all manipulation functions
     * involving the created cursor handle is performed on the bound
     * database.
     *
     * @param cursor              Location for returned cursor handle.
     * @param db                  Database handle.
     * @return                    Standard error code.
     */
    ibret IBCALL IBCursorCreate(ibcursorhandle *cursor, ibdbhandle db);

    /** Destroy cursor handle.
     * Destroys an existing cursor handle and removes all information
     * associated with the cursor handle, e.g. current field, current
     * record. This operation does not change any information
     * in the database.
     *
     * @param cursor              Cursor handle.
     * @return                    Standard error code.
     */
    ibret IBCALL IBCursorDestroy(ibcursorhandle cursor);

    /** Fetch data from current field.
     * Fetches data and data attributes from the current field of the
     * current record.
     *
     * Most of the parameters may be ignored by passing NULL pointers
     * instead of pointers to destination locations for the data. If
     * possible, these data items will not be retrieved from the
     * buffers, thus improving the performance of the function. In clear
     * text: If you don't need a value, pass a NULL pointer; don't pass
     * a temporary variable and discard it.
     *
     * If the actual size of the data which should be fetched exceeds
     * the size of the buffer provided by the user, the function will
     * return an error message and will not change the contents of the
     * buffer.
     *
     * @param cursor              Cursor handle.
     * @param type                Location for the type of the 
     *                            current record field. May be ignored
     *                            by passing a NULL pointer, in which
     *                            case the type will not be retrieved.
     * @param data                Pointer to a data buffer into which
     *                            the actual data of the current
     *                            field should be copied. May be 
     *                            ignored by passing a NULL, in which
     *                            case data will not be retrieved from
     *                            the instance if possible. If a valid
     *                            pointer is provided, the size of the
     *                            data buffer must be passed 
     *                            in the parameter 'n' for all data
     *                            types. 
     * @param n                   Size of data buffer 'data'.
     * @param params              Pointer to a user provided structure
     *                            of parameters modifying the
     *                            behaviour of the data copying
     *                            operation. All members of the
     *                            structure have default values
     *                            which may be passed by 
     *                            passing a NULL pointer. The same
     *                            effect is achived by zeroing all
     *                            members of the structure. Note that
     *                            if a valid pointer to the 
     *                            structure is passed, all member
     *                            values have to be set to valid 
     *                            values, regardless of the
     *                            applicability to the actual data
     *                            being retrieved by the function.
     * @param actsize             Location for the actual data size
     *                            of the current record
     *                            field. May be ignored by passing a
     *                            NULL pointer.
     * @param fielddba            Location for the database address of
     *                            the current record
     *                            field. May be ignored by passing a 
     *                            NULL pointer.
     * @param ownerdba            Location for the database address of
     *                            the owner of the current
     *                            record field. May be ignored by
     *                            passing a NULL pointer.
     * @return                    Standard error code. 
     *                            IBERR_CURSOR_SUPPLIED_FETCH_BUFFER_TOO_SMALL 
     *                            is returned if the size of the read
     *                            data exceeds the size of the data
     *                            buffer provided. 
     */    
    ibret IBCALL IBCursorFetchField(ibcursorhandle cursor, ibdcttypehandle *type,
                                    void *data, ibobjsize n,
                                    struct ibdtparams *params, 
                                    ibobjsize *actsize, ibdbaddr *fielddba, 
                                    ibdbaddr *ownerdba);

    /** Append a new field to the current record.
     * Appends a new field to the record at the position
     * determined by the state of the cursor handle. All properties of
     * the new field must be provided.
     *
     * @param cursor              Cursor handle.
     * @param type                Type of new field.
     * @param data                Pointer to buffer containing the
     *                            data of
     *                            the new field. The size of the
     *                            data buffer must be passed 
     *                            in the parameter 'n'.
     * @param n                   Size of data buffer 'data'. The
     *                            value is ignored if the base type of
     *                            the data being stored is any 
     *                            other than IBBASE_TYPE_BLOB.
     * @param params              Pointer to a user provided structure
     *                            of parameters modifying the
     *                            behaviour of the data copying
     *                            operation. All members of the
     *                            structure have default values
     *                            which may be passed by 
     *                            passing a NULL pointer. The same
     *                            effect is achived by zeroing all
     *                            members of the structure. Note that
     *                            if a valid pointer to the 
     *                            structure is passed, all member
     *                            values have to be set to valid 
     *                            values, regardless of the
     *                            applicability to the actual data
     *                            being retrieved by the function.
     * @return                    Standard error code.
     * @see                       IBCursorNewRecord
     * @see                       IBCursorModifyField
     */
    ibret IBCALL IBCursorAppendField(ibcursorhandle cursor, ibdcttypehandle type,
                                     const void *data, ibobjsize n, 
                                     struct ibdtparams *params);

    /** Modify the current field.
     * Modifies the current field in the record. The previous
     * data is overwritten and other necessary alterations are made to
     * the attributes of the field.
     *
     * This function can be time-consuming depending on the
     * combination of data contents and data attributes which is
     * changed. In general, changing the base type of a field or
     * increasing the size of the data contents of the field are the
     * most costly operations.
     * 
     * @param cursor              Cursor handle.
     * @param type                Pointer to the new type of the
     *                            field, or NULL if the type
     *                            should not be changed.  
     * @param data                Pointer to buffer containing the
     *                            data of the field, or NULL if the
     *                            data should not be changed. If 
     *                            a valid pointer is passed, the size
     *                            of the data buffer must be passed in
     *                            the parameter 'n'.  
     * @param n                   Size of data buffer 'data'. If the
     *                            data buffer pointer is NULL, this
     *                            parameter is ignored whatever value.
     * @param params              Pointer to a user provided structure
     *                            of parameters modifying the
     *                            behaviour of the data copying
     *                            operation. All members of the
     *                            structure have default values
     *                            which may be passed by 
     *                            passing a NULL pointer. The same
     *                            effect is achived by zeroing all
     *                            members of the structure. Note that
     *                            if a valid pointer to the 
     *                            structure is passed, all member
     *                            values have to be set to valid 
     *                            values, regardless of the
     *                            applicability to the actual data
     *                            being retrieved by the function.
     * @return                    Standard error code.
     * @see                       IBCursorAppendField
     */
    ibret IBCALL IBCursorModifyField(ibcursorhandle cursor, 
                                     ibdcttypehandle *type,
                                     const void *data, ibobjsize n,
                                     struct ibdtparams *params);

    /** Delete the current field.
     * Deletes the current field from the current record.
     *
     * When the field is deleted, the current field pointer is moved
     * to the field preceeding the deleted field.
     * 
     * @param cursor              Cursor handle.
     * @return                    Standard error code.
     * @see                       IBCursorEraseRecord
     */
    ibret IBCALL IBCursorEraseField(ibcursorhandle cursor);

    /** Delete the current record.
     * Deletes the entire current record.
     *
     * When the record is deleted, the state information of the cursor
     * handle is undefined. 
     * 
     * @param cursor              Cursor handle.
     * @return                    Standard error code.
     * @see                       IBCursorEraseField
     */
    ibret IBCALL IBCursorEraseRecord(ibcursorhandle cursor);

    /** Create a new record.
     * Creates a new record and changes the current field pointer to
     * an undefined state. A subsequent call to IBCursorAppendField()
     * will write the first field of the record.
     *
     * @param cursor              Cursor handle.
     * @return                    Standard error code. 
     * @see                       IBCursorAppendField
     */
    ibret IBCALL IBCursorNewRecord(ibcursorhandle cursor);

    /** Move the current field pointer.
     * Moves the current field pointer to a field in the current
     * record indexed by a number in the range from 0 to the number 
     * returned by IBCursorGetNumFields() - 1. 
     *
     * Note that when a field is deleted from a record, the
     * indices of the fields in the record will be altered.
     *
     * If the call fails, the current record and  current field
     * pointer of the cursor handle remains unaltered.
     *
     * @param cursor              Cursor handle.
     * @return                    Standard error code.
     * @see                       IBCursorGetNumFields
     * @see                       IBCursorGetFieldIndex
     */
    ibret IBCALL IBCursorGotoField(ibcursorhandle cursor, ibir index);

    /** Return the index of the current field.
     * Returns the index of the current field into the current
     * record. The index is in the range from 0 to the number returned
     * by IBCursorGetNumFields() - 1.  
     *
     * @param cursor              Cursor handle.
     * @param index               Location for returned field index.
     * @return                    Standard error code.
     * @see                       IBCursorGotoField
     */
    ibret IBCALL IBCursorGetFieldIndex(ibcursorhandle cursor, ibir *index);

    /** Get the number of fields in the current record.
     * Gets the number of fields in the current record.
     *
     * Does not modify the current record or current field pointers of
     * the cursor handle.
     *
     * @param cursor              Cursor handle.
     * @param idx                 Location of a variable into which to
     *                            put the number of fields in the
     *                            current record. 
     * @return                    Standard error code.
     */
    ibret IBCALL IBCursorGetNumFields(ibcursorhandle cursor, ibir *num);

    /** Move the current field pointer to a database address.
     * Moves the current field pointer to a database address. The
     * field may or may not be in the current record. All
     * state information in the cursor handle is changed by this
     * operation. 
     *
     * @param cursor              Cursor handle.
     * @return                    Standard error code.
     */
    ibret IBCALL IBCursorGotoMemberAddress(ibcursorhandle cursor, 
                                           ibdbaddr dba);

    /* Query functions. */
    /* ================ */

    /** Create new query expression.
     * Creates a new query expression.
     *
     * Note that a query expression has to be destroyed after use
     * using the function IBQueryDestroy.
     *
     * @param q                   Location for the new query
     *                            expression. 
     * @param db                  Database handle of the database from
     *                            which the data set should be selected.
     * @return                    Standard error code.
     * @see                       IBQueryDestroy
     */
    ibret IBCALL IBQueryCreate(ibqueryhandle *q, ibdbhandle db);

    /** Execute a query.
     * Executes a query given by a query expression and returns a
     * result set containing the data set which matches the query
     * expression. 
     *
     * All duplicate fields are eliminated from the returned set.
     *
     * Note that the returned result set should be destroyed using
     * the function IBResDestroy after use.
     *
     * @param q                   Query expression to be used for the
     *                            data set selection.
     * @param res                 Location for a result set containing
     *                            the result of the data set selection.
     * @return                    Standard error code.
     * @see                       IBQueryDestroy
     * @see                       IBResDestroy
     */
    ibret IBCALL IBQueryExecute(ibqueryhandle q, ibreshandle *res);

    /** Get a pointer to the query parameters of a query handle. 
     * Returns a pointer to the query parameters of a query handle.
     *
     * @param q                   Query handle of node.
     * @param params              Location for pointer to query
     *                            parameter structure.
     * @return                    Standard error code.
     */
    ibret IBCALL IBQueryGetParams(ibqueryhandle q, struct ibqueryparams **params);

    /** Destroy a query expression.
     * Destroys a query expression and optionally all sub-expression associated
     * with the expression.
     *
     * @param q                   Query expression to be destroyed.
     * @param flags               If set to IBQUERY_DESTROY_RECURSIVE,
     *                            will recurse through the query tree
     *                            and destroy all sub-expressions. If
     *                            set to
     *                            IBQUERY_DESTROY_NON_RECURSIVE, will
     *                            destroy only this query handle.
     * @return                    Standard error code.
     * @see                       IBQueryCreate
     */
    ibret IBCALL IBQueryDestroy(ibqueryhandle q, ibquerydestroyflags flags);

    /* Result set functions. */
    /* ===================== */

    /** Destroy a result set.
     * Destroys a result set.
     *
     * @param res                 Result set to be destroyed.
     * @return                    Standard error code.
     */
    ibret IBCALL IBResDestroy(ibreshandle res);

    /** Merge two result sets.
     * Merges two result sets by applying a binary set operator to the
     * two sets. The merged result set replaces the result set on the 
     * left of the binary operator.
     *
     * All duplicate fields are eliminated from the returned set.
     * 
     * @param r                   Result set on the left of the binary
     *                            operator which is replaced by the
     *                            merged result set. 
     * @param s                   Result set on the right of the
     *                            binary operator.
     * @param op                  Binary set operator.
     * @return                    Standard error code.
     * @see                       IBResClone
     */
    ibret IBCALL IBResMerge(ibreshandle r, ibreshandle s, ibsetbinop op);

    /** Clone a result set.
     * Clones a result set and applies a unary set operator to the
     * operation.
     *
     * @param new                 Location for the new result set.
     * @param old                 Result set to be cloned.
     * @param op                  Unary set operation to be applied.
     * @return                    Standard error code.
     * @see                       IBResMerge
     */
    ibret IBCALL IBResClone(ibreshandle *cloned, ibreshandle old, 
                            ibsetunop op);

    /* Result list functions. */
    /* ====================== */

    /** Create result list.
     * Creates a result list from a result set and optionally sorts
     * the result list based on the data content of one or more
     * fields. 
     *
     * @param retlist             Location for the new result list
     *                            handle. 
     * @param resset              Result set which should be
     *                            transformed to a result list and
     *                            optionally sorted. 
     * @param num_sort_steps      Number of sorting steps. 0 if no
     *                            sorting should take place.
     * @param sort_steps          Pointer to an array containing the
     *                            sorting steps. NULL if no sorting
     *                            should take place. 
     * @return                    Standard error code.
     */
    ibret IBCALL IBResListCreate(ibreslisthandle *reslist, 
                                 ibreshandle resset,
                                 unsigned long num_sort_steps,
                                 struct ibreslistsortstep *sort_steps);

    /** Get a result item from a result list.
     * Gets a result item from a result list and sets the state
     * information in a cursor handle to point to the result item, 
     * i.e. the current field pointer of the cursor handle will point to
     * the result element retrieved from the result list.
     * 
     * The result element is chosen from the result list using an
     * index. To retrieve all elements from a result list, the index
     * into the result list may be iterated from 0 to the value
     * returned by IBResGetSize() - 1. 
     *
     * Note that if the result list has not been sorted, the result list
     * elements are in an unpredictable order. 
     *
     * @param res                 Result list from which an item should
     *                            be chosen.
     * @param cursor              Cursor handle which should be pointed
     *                            at the chosen field in the result
     *                            list. 
     * @param item                Index of item in the result list.
     * @return                    Standard error code.
     * @see                       IBResListGetSize
     */
    ibret IBCALL IBResListGetItem(ibreslisthandle res, 
                                  ibcursorhandle cursor, 
                                  ibrl item);

    /** Get the number of elements in a result list.
     * Returns the number of elements in a result list.
     *
     * @param res                 Result list from which the number of
     *                            elements should be returned.
     * @aparm items               Location of variable into which the
     *                            number of elements should be returned.
     * @return                    Standard error code.
     */
    ibret IBCALL IBResListGetSize(ibreslisthandle res, 
                                  ibrl *items);

    /** Destroy a result list.
     * Destroys a result list.
     *
     * @param res                 Result list to be destroyed.
     * @return                    Standard error code.
     */
    ibret IBCALL IBResListDestroy(ibreslisthandle res);

    /* Dictionary functions. */
    /* ===================== */

    /** Create a new or open an existing dictionary.
     * Creates a new dictionary or opens an existing dictionary if a
     * dictionary already exists in the database.
     *
     * @param dct                 Location of the new dictionary
     *                            handle. 
     * @param db                  Database handle of database to
     *                            associate with the new dictionary.
     * @return                    Standard error code.
     * @see                       IBDctDestroy
     */
    ibret IBCALL IBDctCreate(ibdcthandle *dct, ibdbhandle db);

    /** Destroy a dictionary handle.
     * Destroys a dictionary handle and writes all dictionary
     * information to the database.
     *
     * @param dct                 Dictionary handle to destroy.
     * @return                    Standard error code.
     */
    ibret IBCALL IBDctDestroy(ibdcthandle dct);

    /** Create a new type.
     * Creates a new type in a dictionary.
     *
     * @param dct                 Dictionary handle of dictionary to
     *                            use.
     * @param b                   Base type of new type.
     * @param s                   Storage type of new type.
     * @param type                Location for new type handle.
     * @param name                Name of new type. May be
     *                            omitted by passing a NULL pointer.
     * @param enc                 Encoding of the type name. Ignored
     *                            if no type name is provided.
     * @return                    Standard error code.
     */
    ibret IBCALL IBDctNewType(ibdcthandle dct, ibbtype b, ibstype s,
                              ibdcttypehandle *type, const ibechar *name,
                              ibenc enc);

    /** Get type attributes.
     * Gets the attributes of a type from a dictionary.
     *
     * @param dct                 Dictionary handle of dictionary to
     *                            use.
     * @param type                Type handle of the type whose attributes
     *                            should be retrieved.
     * @param b                   Location for the base type of the
     *                            type. May be ignored by passing
     *                            a NULL pointer.
     * @param s                   Location for the storage type of the
     *                            type. May be ignored by passing
     *                            a NULL pointer.
     * @param u                   Location for the usertype index of the
     *                            type. May be ignored by passing
     *                            a NULL pointer.
     * @param name                Location for the name of user
     *                            type. May be ignored by passing a
     *                            NULL pointer. 
     * @param n                   Size of the 'name' buffer in
     *                            bytes. Ignored if the 'name'
     *                            pointer is a NULL pointer. 
     * @param enc                 Encoding of the returned user type
     *                            name. Ignored 
     *                            if no user type name is provided.
     * @return                    Standard error code.
     */
    ibret IBCALL IBDctGetTypeAttributes(ibdcthandle dct, 
                                        ibdcttypehandle type, 
                                        ibbtype *b, ibstype *s, ibutype *u,  
                                        ibechar *name, ibobjsize n,
                                        ibenc enc); 

    /** Delete a type.
     * Deletes a type from a dictionary. This deletes the 
     * type's entry in the dictionary, and thus removes the
     * association between a type, a type name, a specific
     * base type and a specific storage type. However, it does not
     * change any of the data in the database. If IBDctNewType()
     * is called later, the new type may assume the identity of a
     * deleted type.
     *
     * @param dct                 Dictionary handle of dictionary to
     *                            use.
     * @param type                Type handle of type to delete.
     * @return                    Standard error code.
     */
    ibret IBCALL IBDctDeleteType(ibdcthandle dct, ibdcttypehandle type);

    /** List types.
     * Lists types in a dictionary.
     *
     * @param dct                 Dictionary handle of dictionary to
     *                            use.
     * @param list                A pointer to a user-provided array
     *                            into which the type handles of all
     *                            types in the dictionary's namespace
     *                            will be copied, or NULL to ignore. 
     * @param listsize            Number of elements in the list
     *                            of types.
     * @param available           Location for a variable containing
     *                            the number of type handles available 
     *                            for listing. Ignored if NULL.
     * @return                    Standard error code.
     */
    ibret IBCALL IBDctListTypes(ibdcthandle dct, ibdcttypehandle *list,
                                unsigned long listsize, 
                                unsigned long *available);

    /** Look up a type name.
     * Looks up a symbolic type name in a dictionary and returns
     * the type handle of the type.
     *
     * @param dct                 Dictionary handle of dictionary to
     *                            use.
     * @param name                Symbolic name of the type.
     * @param type                Location for the type handle of the
     *                            type. 
     * @param enc                 Encoding of the string in 'uname'.
     * @return                    Standard error code.
     */
    ibret IBCALL IBDctLookupName(ibdcthandle dct, 
                                 const ibechar *name,
                                 ibdcttypehandle *type, ibenc enc); 

    /* Error handling functions. */
    /* ========================= */

    /** Get error message corresponding to an error code.
     * Returns an error message corresponding the an error code. The
     * error message includes both the error message produced by the 
     * database system and an error code from the operating system, if
     * any. 
     *
     * @param code                Standard error code.
     * @return                    Returns a pointer to a zero-terminated
     *                            string containing a description of the
     *                            error code. The string will be
     *                            overwritten by subsequent calls to
     *                            this functions.
     */
    const ibtchar IBCALL *IBErrorMessage(ibret code);

#ifdef __cplusplus
};
#endif /* __cplusplus */

#endif /* __BDE_H__ */

/* 
   Local Variables:
   mode:c
   indent-tabs-mode:nil
   c-basic-offset:4
   End:
*/
